__title__ = "jadoc"
__description__ = "Tokenizes Japanese documents to enable CRUD operations."
__url__ = "https://github.com/poyo46/jadoc"
__version__ = "0.2.4"
__author__ = "poyo46"
__author_email__ = "poyo4rock@gmail.com"
__license__ = "Apache-2.0"
__copyright__ = "Copyright 2021 poyo46"
